/*
    Fingerprint selection algorithms for local text reuse detection
    Copyright (C) 2020 Gints Jekabsons, gints.jekabsons@rtu.lv
    Available at: http://www.cs.rtu.lv/jekabsons/nlp.html
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
	
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.
	
    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

package lv.rtu.fingerprinting;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import lv.rtu.fingerprinting.Configuration.Algorithm;
import opennlp.tools.ngram.NGramGenerator;
import opennlp.tools.stemmer.PorterStemmer;

public class Document {
	private File file;
	private List<String> allTokens;
	private List<HashPos> hashes;
	private int numAllNGrams;
	private int numFingerprints;
	private int numUniqueFingerprints;
	private HashMap<Integer, String> hashToNGram;
	
	private static final String[] stopwords = { // stop words from https://gist.github.com/sebleier/554280
			"a", "about", "above", "after", "again", "against", "ain", "all", "am", "an", "and", "any", "are", "aren", "aren't", "as", "at", "be",
			"because", "been", "before", "being", "below", "between", "both", "but", "by", "can", "couldn", "couldn't", "d", "did", "didn", "didn't",
			"do", "does", "doesn", "doesn't", "doing", "don", "don't", "down", "during", "each", "few", "for", "from", "further", "had", "hadn",
			"hadn't", "has", "hasn", "hasn't", "have", "haven", "haven't", "having", "he", "her", "here", "hers", "herself", "him", "himself", "his",
			"how", "i", "if", "in", "into", "is", "isn", "isn't", "it", "it's", "its", "itself", "just", "ll", "m", "ma", "me", "mightn", "mightn't",
			"more", "most", "mustn", "mustn't", "my", "myself", "needn", "needn't", "no", "nor", "not", "now", "o", "of", "off", "on", "once", "only",
			"or", "other", "our", "ours", "ourselves", "out", "over", "own", "re", "s", "same", "shan", "shan't", "she", "she's", "should", "should've",
			"shouldn", "shouldn't", "so", "some", "such", "t", "than", "that", "that'll", "the", "their", "theirs", "them", "themselves", "then",
			"there", "these", "they", "this", "those", "through", "to", "too", "under", "until", "up", "ve", "very", "was", "wasn", "wasn't", "we",
			"were", "weren", "weren't", "what", "when", "where", "which", "while", "who", "whom", "why", "will", "with", "won", "won't", "wouldn",
			"wouldn't", "y", "you", "you'd", "you'll", "you're", "you've", "your", "yours", "yourself", "yourselves", "could", "he'd", "he'll", "he's",
			"here's", "how's", "i'd", "i'll", "i'm", "i've", "let's", "ought", "she'd", "she'll", "that's", "there's", "they'd", "they'll", "they're",
			"they've", "we'd", "we'll", "we're", "we've", "what's", "when's", "where's", "who's", "why's", "would"};
	
	/**
	 * Reads the file, tokenizes it, optionally removes stop words, stored the list of words (tokens).
	 * @param file - the file of the document
	 * @param config - the configuration from which only the <code>removeStopWords</code> is used
	 * @throws IOException - thrown in case the file can't be properly read
	 */
	public Document(File file, Configuration config) throws IOException {
		this.file = file;
		if (config.verbose >= 2)
			System.out.println("Reading file \"" + file.getName() + "\"");
		
		String text = new String(Files.readAllBytes(file.toPath()), "UTF-8");
		text = text.replaceAll("[^\\p{IsAlphabetic}\\p{Digit}]", " ").replaceAll(" +", " ").trim().toLowerCase();
		
		allTokens = new ArrayList<String>(Arrays.asList(text.split(" ")));
		if (config.removeStopWords)
			allTokens.removeAll(Arrays.asList(stopwords));
	}
	
	/**
	 * Creates fingerprints according to the provided configuration
	 * @param config - the configuration.
	 * @param hashFreq - n-gram frequencies. Used by FBW and MFBW algorithms or filled (see next parameter).
	 * @param fillHashFreq - whether to compute n-gram frequencies and add them to those in the hashFreq object.
	 */
	public void createFingerprints(Configuration config, HashMap<Integer, Integer> hashFreq, boolean fillHashFreq) {
		if (config.verbose >= 2)
			System.out.println("Processing file \"" + file.getName() + "\"");
		
		List<String> nGrams = getListOfNGrams(config);
		
		if (config.verbose >= 1) {
			hashToNGram = new HashMap<Integer, String>();
			if (config.verbose >= 2)
				System.out.println("\"" + file.getName() + "\" list (" + nGrams.size() + "): " + nGrams.toString());
		}
		
		numAllNGrams = nGrams.size();
		
		hashes = new ArrayList<HashPos>(nGrams.size());
		for (int i = 0; i < nGrams.size(); i++) {
			String nGram = nGrams.get(i);
			
			int hash = computeHash(nGram);
			hashes.add(new HashPos(hash, nGram));
			
			if (config.verbose >= 1)
				hashToNGram.put(hash, nGram);
		}
		
		switch (config.algorithm) {
			case EVERY_PTH:
				hashes = everyPth(hashes, config.parameter);
				break;
			case MODP:
				hashes = modp(hashes, config.parameter);
				break;
			case WINNOWING:
				hashes = winnowing(hashes, config.parameter);
				break;
			case FBW:
				hashes = freqWinnowing(hashes, config.parameter, hashFreq, false);
				break;
			case MFBW:
				hashes = freqWinnowing(hashes, config.parameter, hashFreq, true);
				break;
			default: // for FULL and HAILSTORM
		}
		
		numFingerprints = hashes.size();
		Set<HashPos> set = new HashSet<HashPos>(hashes);
		numUniqueFingerprints = set.size();
		
		if (fillHashFreq && (hashFreq != null)) {
			for (HashPos hashPos : hashes) {
				Integer freq = hashFreq.get(hashPos.hash);
				if (freq == null)
					hashFreq.put(hashPos.hash, 1);
				else
					hashFreq.put(hashPos.hash, freq + 1);
			}
		}
		
		hashes = new ArrayList<HashPos>(set); // we will use set instead of multiset because we don't need the repetitions
		
		Collections.sort(hashes); // here we sort the hashes so that later overlap computation could be made much faster
		
		if (config.verbose >= 2)
			System.out.println("\"" + file.getName() + "\" list (" + hashes.size() + "): " + hashes.toString());
	}
	
	private List<String> getListOfNGrams(Configuration config) {
		PorterStemmer stemmer = null;
		if (config.doStemming)
			stemmer = new PorterStemmer(); //create new stemmer because the PorterStemmer class is not thread-safe
		
		List<String> usedTokens = new ArrayList<String>(allTokens.size());
		for (String token : allTokens) {
			if (config.doStemming)
				token = stemmer.stem(token);
			if (token.length() < config.minWordLength)
				continue;
			usedTokens.add(token);
		}
		
		if (config.nGramsConsistOfWords) {
			if (config.algorithm == Algorithm.HAILSTORM)
				return generateHailstormNGrams(usedTokens, config.gramSize, " ");
			else
				return NGramGenerator.generate(usedTokens, config.gramSize, " ");
		} else {
			char[] chars = String.join("", usedTokens).toCharArray();
			if (config.algorithm == Algorithm.HAILSTORM)
				return generateHailstormNGrams(chars, config.gramSize);
			else
				return NGramGenerator.generate(chars, config.gramSize, "");
		}
	}
	
	// https://dl.acm.org/doi/pdf/10.1145/1526709.1526719
	private List<String> generateHailstormNGrams(List<String> tokens, int n, String separator) {
		List<Integer> hashes = new ArrayList<Integer>(tokens.size());
		for (String token : tokens)
			hashes.add(computeHash(token));
		List<String> outputNGrams = new ArrayList<>();
		for (int i = 0; i < tokens.size() - n + 1; i++) {
			StringBuilder sb = new StringBuilder();
			int minHashIndex = i;
			for (int x = i; x < i + n; x++) {
				if (x > i)
					sb.append(separator);
				sb.append(tokens.get(x));
				if (hashes.get(x) < hashes.get(minHashIndex))
					minHashIndex = x;
			}
			if ((hashes.get(minHashIndex) == hashes.get(i)) || (hashes.get(minHashIndex) == hashes.get(i + n - 1)))
				outputNGrams.add(sb.toString());
		}
		return outputNGrams;
	}
	
	// https://dl.acm.org/doi/pdf/10.1145/1526709.1526719
	private List<String> generateHailstormNGrams(char[] chars, int n) {
		List<Integer> hashes = new ArrayList<Integer>(chars.length);
		for (char c : chars)
			hashes.add(computeHash(Character.toString(c)));
		
		List<String> outputNGrams = new ArrayList<>();
		for (int i = 0; i < chars.length - n + 1; i++) {
			StringBuilder sb = new StringBuilder();
			int minHashIndex = i;
			for (int x = i; x < i + n; x++) {
				sb.append(chars[x]);
				if (hashes.get(x) < hashes.get(minHashIndex))
					minHashIndex = x;
			}
			if ((hashes.get(minHashIndex) == hashes.get(i)) || (hashes.get(minHashIndex) == hashes.get(i + n - 1)))
				outputNGrams.add(sb.toString());
		}
		return outputNGrams;
	}
	
	private List<HashPos> everyPth(List<HashPos> input, int p) {
		List<HashPos> output = new ArrayList<HashPos>(input.size() / p + 1);
		for (int index = 0; index < input.size(); index += p)
			output.add(input.get(index));
		return output; 
	}
	
	// http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.12.3222
	private List<HashPos> modp(List<HashPos> input, int p) {
		List<HashPos> output = new ArrayList<HashPos>();
		for (HashPos hash : input)
			if (hash.hash % p == 0)
				output.add(hash);
		return output; 
	}
	
	// https://homepages.warwick.ac.uk/~masgar/Maths/winnowing.pdf
	// https://homepages.warwick.ac.uk/~masgar/Maths/winnowing.html
	private List<HashPos> winnowing(List<HashPos> input, int w) {
		int n = input.size();
		if (w > n)
			w = n;
		List<HashPos> output = new ArrayList<HashPos>();
		int lastMinIndex = -1;
		int minHash = Integer.MAX_VALUE;
		for (int i = 0; i < n - w + 1; i++) {
			if (i <= lastMinIndex) {
				int j = i + w - 1;
				if (input.get(j).hash <= minHash) {
					minHash = input.get(j).hash;
					output.add(input.get(j));
					lastMinIndex = j;
				}
			} else {
				minHash = Integer.MAX_VALUE;
				int minIndex = i + w - 1;
				for (int j = minIndex; j >= i; j--) {
					if (input.get(j).hash < minHash) {
						minHash = input.get(j).hash;
						minIndex = j;
					}
				}
				if (lastMinIndex != minIndex) {
					output.add(input.get(minIndex));
					lastMinIndex = minIndex;
				}
			}
		}
		return output;
	}
	
	// http://qinjianbin.com/files/wise2013-sun.pdf
	// https://content.sciendo.com/view/journals/acss/25/1/article-p11.xml
	private List<HashPos> freqWinnowing(List<HashPos> input, int w, HashMap<Integer, Integer> hashFreq, boolean MFBW) {
		int size = input.size();
		if (w > size)
			w = size;
		List<HashPos> output = new ArrayList<HashPos>();
		int lastMinIndex = -1;
		for (int i = 0; i < size - w + 1; i++) {
			int minIndex = i + w - 1;
			HashPos savedHash = input.get(minIndex);
			Integer freq = hashFreq.get(savedHash.hash);
			int minFreq;
			if (freq == null) {
				if (MFBW)
					minFreq = Integer.MAX_VALUE;
				else
					minFreq = 0;
			} else {
				minFreq = freq;
			}
			for (int j = minIndex - 1; j >= i; j--) {
				HashPos candidateHash = input.get(j);
				freq = hashFreq.get(candidateHash.hash);
				if (freq == null) {
					if (MFBW)
						freq = Integer.MAX_VALUE;
					else
						freq = 0;
				}
				if ((freq < minFreq) ||
					(
						(freq == minFreq) && (candidateHash.nGram.compareTo(savedHash.nGram) < 1) // use alphabet to decide
						//(freq == minFreq) && (candidateHash.hash < savedHash.hash) // use hash to decide
					)) {
					savedHash = candidateHash;
					minIndex = j;
					minFreq = freq;
				}
			}
			if (lastMinIndex != minIndex) {
				output.add(savedHash);
				lastMinIndex = minIndex;
			}
		}
		return output;
	}
	
	public File getFile() {
		return file;
	}
	
	public int getNumAllTokens() {
		return allTokens.size();
	}
	
	public int getNumAllNGrams() {
		return numAllNGrams;
	}
	
	public int getNumFingerprints() {
		return numFingerprints;
	}
	
	public int getNumUniqueFingerprints() {
		return numUniqueFingerprints;
	}
	
	/**
	 * Checks whether this document has at least <code>atLeast</code> overlapping n-grams with the <code>document2</code>.
	 * @param document2 - the second document.
	 * @param atLeast - the minimum required overlaps.
	 * @param printOverlappingNGrams - whether to print the overlapping n-grams.
	 * @return Returns <code>true</code> if the minimum overlaps is reached. Otherwise, returns <code>false</code>.
	 */
	public boolean hasOverlaps(Document document2, int atLeast, boolean printOverlappingNGrams) {
		List<HashPos> hashes2 = document2.hashes;
		int numOverlaps = 0;
		int i = 0;
		int j = 0;
		while ((i < hashes.size()) && (j < hashes2.size())) {
			int hash1 = hashes.get(i).hash;
			int hash2 = hashes2.get(j).hash;
			if (hash1 > hash2) {
				j++;
			} else
				if (hash1 < hash2) {
					i++;
				} else {
					if (printOverlappingNGrams && (hashToNGram != null)) {
							System.out.println("Overlap: \"" + file.getName() + "\" & \"" + document2.file.getName() + "\": " +
									hash1 + ", " + hash2 + ", " +
									hashToNGram.get(hash1) + ", " +  document2.hashToNGram.get(hash2));
					}
					numOverlaps++;
					if (numOverlaps >= atLeast)
						return true;
					i++;
					j++;
					// keep going through repeated hash value in case it is a multiset
					for (; i < hashes.size(); i++)
						if (hashes.get(i).hash != hash1)
							break;
					for (; j < hashes2.size(); j++)
						if (hashes2.get(j).hash != hash1)
							break;
				}
		}
		return numOverlaps >= atLeast;
	}
	
	private class HashPos implements Comparable<HashPos> {
		public int hash;
		public String nGram;
		public HashPos(int hash, String nGram) {
			this.hash = hash;
			this.nGram = nGram;
		}
		@Override
		public int compareTo(HashPos other) {
			if (hash < other.hash)
				return -1;
			if (hash > other.hash)
				return 1;
			return 0;
		}
		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if ((obj == null) || (obj.getClass() != this.getClass()))
				return false;
			return ((HashPos) obj).hash == this.hash;
		}
		@Override
		public int hashCode() {
			return hash;
		}
	}
	
	// FNV-1a 32-bit hashing algorithm
	// https://en.wikipedia.org/wiki/Fowler%E2%80%93Noll%E2%80%93Vo_hash_function
	// https://github.com/prasanthj/hasher/blob/master/src/main/java/hasher/FNV1a.java
	private static final int FNV_32_INIT = 0x811c9dc5;
	private static final int FNV_32_PRIME = 0x01000193;
	private static int computeHash(String string) {
		byte[] bytes = string.getBytes(StandardCharsets.UTF_8);
		int rv = FNV_32_INIT;
		for (int i = 0; i < bytes.length; i++) {
			rv ^= (bytes[i] & 0xFF);
			rv *= FNV_32_PRIME;
		}
		return rv;
	}
}