/*
    Fingerprint selection algorithms for local text reuse detection
    Copyright (C) 2020 Gints Jekabsons, gints.jekabsons@rtu.lv
    Available at: http://www.cs.rtu.lv/jekabsons/nlp.html
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
	
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.
	
    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

package lv.rtu.fingerprinting;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * <p>
 * <h1>Fingerprint selection algorithms for local text reuse detection</h1>
 * </p>
 * <p>
 * Developed by Gints Jekabsons.
 * <br>
 * Available at: <a href="http://www.cs.rtu.lv/jekabsons/nlp.html">http://www.cs.rtu.lv/jekabsons/nlp.html</a>.
 * </p>
 * <p>
 * This software was used for the experiments in the paper
 * "Evaluation of Fingerprint Selection Algorithms for Local Text Reuse Detection"
 * (available: <a href="http://www.cs.rtu.lv/jekabsons/Files/Jek_ACSS2020.pdf">http://www.cs.rtu.lv/jekabsons/Files/Jek_ACSS2020.pdf</a> and <a href="https://content.sciendo.com/view/journals/acss/25/1/article-p11.xml">https://content.sciendo.com/view/journals/acss/25/1/article-p11.xml</a>).
 * <b>If you are using this software, please give a reference the paper.</b> And it would be nice to give a link to the website as well: <a href="http://www.cs.rtu.lv/jekabsons/nlp.html">http://www.cs.rtu.lv/jekabsons/nlp.html</a>.
 * </p>
 * <p>
 * This software is developed for evaluating the effectiveness of fingerprint selection algorithms for the source retrieval stage of local text reuse detection.
 * It implements the following fingerprint selection algorithms (see the paper for details):
 * <li>Full fingerprinting;</li>
 * <li>Every p-th;</li>
 * <li>0 mod p;</li>
 * <li>Winnowing;</li>
 * <li>Hailstorm;</li>
 * <li>Frequency-Biased Winnowing (FBW);</li>
 * <li>Modified Frequency-Biased Winnowing (MFBW) - proposed in the paper.</li>
 * </p>
 * <br>
 * <p>
 * To use the software, you need a dataset containing collection documents and query documents.
 * You can set the directories for the files using the static variables of the Fingerprinting class.
 * <br>
 * The text files in the "Query" directory (the query documents) should be numbered, i.e., "1.txt", "2.txt", etc.
 * The text files in the "Collection" directory (the collection documents) should be named in the following way:
 * <li>If a collection document is considered a correct match with a query document,
 * its filename should start with the corresponding query document's number followed by underline character followed by any other characters as convenient,
 * e.g., "1_fileA.txt", "1_fileB.txt", "2_fileC.txt" etc. Note that for all other query documents the collection document will be considered a false match.</li>
 * <li>If a collection document is considered a false match with all query documents, its filename should start with character "F", e.g., "F_fileD.txt".</li>
 * </p>
 * <br>
 * <p>
 * Java dependency: Apache OpenNLP (https://opennlp.apache.org/) - used with version 1.9.0
 * </p>
 * <br>
 * @version 1.0 (June 5, 2020)
 * @author Gints Jekabsons, gints.jekabsons@rtu.lv
 */

public class Fingerprinting {
	private static final String DATA_PATH = "G:/Data/";							// Main directory containing data subdirectories
	private static final String COLLECTION_PATH = DATA_PATH + "Collection/";	// Directory for collection documents
	private static final String QUERY_PATH = DATA_PATH + "Query/";				// Directory for query documents to be checked against the collection documents
	
	private static final String RESULT_PATH = "G:/Data/_Results/";				// Directory for storing the results
	
	public static void main(String[] args) throws Exception {
		(new Fingerprinting()).start();
	}
	
	public void start() throws Exception {
		Configuration config = new Configuration(true, 3, true, true, 1, Configuration.Algorithm.FULL, 0, 0);
		
		System.out.println("Getting file lists...");
		
		List<File> queryFiles = getDirectoryFileList(QUERY_PATH, config);
		List<File> collectionFiles = getDirectoryFileList(COLLECTION_PATH, config);
		
		System.out.println("Mapping...");
		
		// Here we prepare a map from query documents to their correct sources in the collection
		Map<File, HashSet<File>> correctPairs = new HashMap<File, HashSet<File>>();
		Map<File, HashSet<File>> incorrectPairs = null; // Currently unused. Should be used if only specific pairs are to be compared, instead of all possible pairs.
		for (File collectionFile : collectionFiles) {
			String collectionDocFilename = collectionFile.getName();
			if (collectionDocFilename.startsWith("F"))
				continue;
			File file = new File(QUERY_PATH + collectionDocFilename.split("_", 2)[0] + ".txt");
			HashSet<File> set = correctPairs.get(file);
			if (set == null)
				set = new HashSet<File>();
			set.add(collectionFile);
			correctPairs.put(file, set);
		}
		
		System.out.println("Preloading documents...");
		
		// If you have too many documents to preload all of them in the memory,
		// just set preloadedQueryDocs and preloadedCollectionDocs to null and comment-out both for-loops.
		
		Map<File, Document> preloadedQueryDocs = new HashMap<File, Document>();
		for (File file : queryFiles)
			preloadedQueryDocs.put(file, new Document(file, config));
		
		Map<File, Document> preloadedCollectionDocs = new HashMap<File, Document>();
		for (File file : collectionFiles)
			preloadedCollectionDocs.put(file, new Document(file, config));
		
		System.out.println("Running experiments...");
		
		OutputStreamWriter outputWriter = new OutputStreamWriter(new FileOutputStream(RESULT_PATH + config.getResultsFileName() + ".txt"), StandardCharsets.UTF_8);
		outputWriter.write(config.getColumnNames());
		outputWriter.write("\tTP\tFP\tTN\tFN\tPrecision\tRecall\tF1\tFall-out");
		outputWriter.write("\tAll tokens\tAll n-grams\tFingerprints\tUnique fingerprints");
		outputWriter.write("\r\n");
		
		for (config.gramSize = 1; config.gramSize <= 10; config.gramSize++) { //XXX set in accordance to your needs
			
			// If n-gram collection frequencies are required, here we compute them in the collection documents by doing "Full fingerprinting"
			HashMap<Integer, Integer> hashFreq = null;
			if (config.needFreq()) {
				Configuration configBase = config.clone();
				configBase.algorithm = Configuration.Algorithm.FULL;
				
				hashFreq = new HashMap<Integer, Integer>();
				if (preloadedCollectionDocs != null) {
					for (Document doc : preloadedCollectionDocs.values())
						doc.createFingerprints(configBase, hashFreq, true);
				} else {
					for (File file : collectionFiles)
						(new Document(file, config)).createFingerprints(configBase, hashFreq, true);
				}
			}
			
			int firstParam;
			int lastParam;
			if ((config.algorithm == Configuration.Algorithm.FULL) ||
				(config.algorithm == Configuration.Algorithm.HAILSTORM)) {
				// We will not loop because these two algorithms do not have the parameter
				firstParam = 0;
				lastParam = 0;
			} else {
				firstParam = 2;
				// With smaller n-grams we want to try larger parameter values
				lastParam = Math.max(8, (8 - config.gramSize) * 8); //XXX set in accordance to your needs
			}
			
			for (config.parameter = firstParam; config.parameter <= lastParam; config.parameter++) {
				
				// Here we compute some statistics and, if the documents are preloaded, they now will have their fingerprints ready
				int numAllTokens = 0;
				int numAllNGrams = 0;
				int numFingerprints = 0;
				int numUniqueFingerprints = 0;
				if (preloadedCollectionDocs != null) {
					for (Document doc : preloadedCollectionDocs.values()) {
						doc.createFingerprints(config, hashFreq, false);
						numAllTokens += doc.getNumAllTokens();
						numAllNGrams += doc.getNumAllNGrams();
						numFingerprints += doc.getNumFingerprints();
						numUniqueFingerprints += doc.getNumUniqueFingerprints();
					}
				} else {
					for (File file : collectionFiles) {
						Document doc = new Document(file, config);
						doc.createFingerprints(config, hashFreq, false);
						numAllTokens += doc.getNumAllTokens();
						numAllNGrams += doc.getNumAllNGrams();
						numFingerprints += doc.getNumFingerprints();
						numUniqueFingerprints += doc.getNumUniqueFingerprints();
					}
				}
				
				if (preloadedQueryDocs != null) {
					for (Document doc : preloadedQueryDocs.values())
						doc.createFingerprints(config, hashFreq, false);
				}
				
				// With smaller n-grams we want to try larger minOverlap values
				int lastMinOverlap;
				if (config.algorithm == Configuration.Algorithm.FULL)
					lastMinOverlap = Math.max(10, (8 - config.gramSize) * 40 - 50); //XXX set in accordance to your needs
				else
					lastMinOverlap = Math.max(7, (7 - config.gramSize) * 7); //XXX set in accordance to your needs
				
				for (config.minOverlapNGrams = 1; config.minOverlapNGrams <= lastMinOverlap; config.minOverlapNGrams++) {
					System.out.println("gramSize=" + config.gramSize + ", parameter=" + config.parameter + ", minOverlapNGrams=" + config.minOverlapNGrams);
					
					String results = analyse(queryFiles, collectionFiles, correctPairs, incorrectPairs, preloadedQueryDocs, preloadedCollectionDocs, hashFreq, config);
					
					outputWriter.write(config.toString());
					outputWriter.write("\t");
					outputWriter.write(results);
					outputWriter.write("\t" + numAllTokens / collectionFiles.size() + "\t" + numAllNGrams / collectionFiles.size() + "\t" + numFingerprints / collectionFiles.size() + "\t" + numUniqueFingerprints / collectionFiles.size());
					outputWriter.write("\r\n");
					outputWriter.flush();
				}
				
			}
		}
		
		outputWriter.close();
		
		System.out.println("Done");
	}
	
	private String analyse(
			List<File> queryDocs,
			List<File> collectionDocs,
			Map<File, HashSet<File>> correctPairs,
			Map<File, HashSet<File>> incorrectPairs,
			Map<File, Document> preloadedQueryDocs,
			Map<File, Document> preloadedCollectionDocs,
			HashMap<Integer, Integer> hashFreq,
			Configuration config
			) throws InterruptedException {
		
		final AtomicInteger queryDocIndex = new AtomicInteger(0);
		final AtomicInteger tp = new AtomicInteger(0);
		final AtomicInteger fp = new AtomicInteger(0);
		final AtomicInteger tn = new AtomicInteger(0);
		final AtomicInteger fn = new AtomicInteger(0);
		final int threadCount = 6;
		ExecutorService executor = Executors.newFixedThreadPool(threadCount);
		ExecutorCompletionService<Integer> completionService = new ExecutorCompletionService<Integer>(executor);
		for (int i = 0; i < threadCount; i++) {
			Callable<Integer> task = () -> {
				check(queryDocs, collectionDocs, correctPairs, incorrectPairs, preloadedQueryDocs, preloadedCollectionDocs, hashFreq, queryDocIndex, tp, fp, tn, fn, config);
				return null;
			};
			completionService.submit(task);
		}
		for (int i = 0; i < threadCount; i++) {
			try {
				completionService.take().get();
			} catch (Exception e) {
				e.printStackTrace();
				executor.shutdownNow();
				break;
			}
		}
		executor.shutdown();
		executor.awaitTermination(7, TimeUnit.DAYS);
		
		double denominator = tp.doubleValue() + fp.doubleValue();
		double precision = denominator > 0 ? tp.doubleValue() / denominator : 1;
		
		denominator = tp.doubleValue() + fn.doubleValue();
		double recall = denominator > 0 ? tp.doubleValue() / denominator : 0;
		
		double f1 = (precision > 0) && (recall > 0) ? 2 * (precision * recall) / (precision + recall) : 0;
		
		denominator = fp.doubleValue() + tn.doubleValue();
		double fallOut = denominator > 0 ? fp.doubleValue() / denominator : 0;
		
		return tp + "\t" + fp + "\t" + tn + "\t" + fn + "\t" + precision + "\t" + recall + "\t" + f1 + "\t" + fallOut;
	}
	
	private void check(
			List<File> queryFiles,
			List<File> collectionFiles,
			Map<File, HashSet<File>> correctPairs,
			Map<File, HashSet<File>> incorrectPairs,
			Map<File, Document> preloadedQueryDocs,
			Map<File, Document> preloadedCollectionDocs,
			HashMap<Integer, Integer> hashFreq,
			AtomicInteger queryFilesIndex,
			AtomicInteger tpSum,
			AtomicInteger fpSum,
			AtomicInteger tnSum,
			AtomicInteger fnSum,
			Configuration config
			) throws IOException {
		
		while (true) {
			if (Thread.currentThread().isInterrupted())
				break;
			
			int i = queryFilesIndex.getAndIncrement();
			if (i >= queryFiles.size())
				break;
			
			int tp = 0;
			int fp = 0;
			int tn = 0;
			int fn = 0;
			File queryFile = queryFiles.get(i);
			
			Document queryDoc = null;
			if (preloadedQueryDocs != null)
				queryDoc = preloadedQueryDocs.get(queryFile);
			if (queryDoc == null) {
				queryDoc = new Document(queryFile, config);
				queryDoc.createFingerprints(config, hashFreq, false);
			}
			
			if (incorrectPairs != null) {
				
				Set<File> collectionFilesToCompare = correctPairs.get(queryFile);
				if (collectionFilesToCompare != null) {
					for (File collectionFile : collectionFilesToCompare) {
						Document collectionDoc = null;
						if (preloadedCollectionDocs != null)
							collectionDoc = preloadedCollectionDocs.get(collectionFile);
						if (collectionDoc == null) {
							collectionDoc = new Document(collectionFile, config);
							collectionDoc.createFingerprints(config, hashFreq, false);
						}
						
						if (queryDoc.hasOverlaps(collectionDoc, config.minOverlapNGrams, false))
							tp++;
						else
							fn++;
					}
				}
				
				collectionFilesToCompare = incorrectPairs.get(queryFile);
				if (collectionFilesToCompare != null) {
					for (File collectionFile : collectionFilesToCompare) {
						Document collectionDoc = null;
						if (preloadedCollectionDocs != null)
							collectionDoc = preloadedCollectionDocs.get(collectionFile);
						if (collectionDoc == null) {
							collectionDoc = new Document(collectionFile, config);
							collectionDoc.createFingerprints(config, hashFreq, false);
						}
						
						if (queryDoc.hasOverlaps(collectionDoc, config.minOverlapNGrams, config.verbose >= 1))
							fp++;
						else
							tn++;
					}
				}
				
			} else {
				
				Set<File> correctSources = correctPairs.get(queryFile);
				for (File collectionFile : collectionFiles) {
					Document collectionDoc = null;
					if (preloadedCollectionDocs != null)
						collectionDoc = preloadedCollectionDocs.get(collectionFile);
					if (collectionDoc == null) {
						collectionDoc = new Document(collectionFile, config);
						collectionDoc.createFingerprints(config, hashFreq, false);
					}
					
					boolean actualPositive = (correctSources != null) ? correctSources.contains(collectionDoc.getFile()) : false;
					if (queryDoc.hasOverlaps(collectionDoc, config.minOverlapNGrams, (config.verbose >= 1) && !actualPositive)) {
						if (actualPositive)
							tp++;
						else
							fp++;
					} else {
						if (actualPositive)
							fn++;
						else
							tn++;
					}
				}
				
			}
			
			tpSum.addAndGet(tp);
			fpSum.addAndGet(fp);
			tnSum.addAndGet(tn);
			fnSum.addAndGet(fn);
		}
	}
	
	private static List<File> getDirectoryFileList(String dirPath, Configuration config) throws IOException {
		File dir = new File(dirPath);
		List<File> files = new ArrayList<File>();
		if (dir.exists()) {
			File[] filesInDir = dir.listFiles();
			for (File file : filesInDir) {
				if (file.isDirectory())
					continue;
				files.add(file);
			}
		}
		return files;
	}
}
